/***************************************************************
 *
 * arch/i386/include/portio.h
 * port I/O functions for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#ifndef ___I386_PORTIO_H_
#define ___I386_PORTIO_H_

#include <stdint.h>

static inline void outb(int16_t port, int8_t byte)
{
	asm volatile(
		  "outb %0, %1"
		:
		: "a" (byte), "Nd" (port)
	);

	return;
};

static inline void outw(int16_t port, int16_t word)
{
	asm volatile(
		  "outw %0, %1"
		:
		: "a"(word), "Nd" (port)
	);

	return;
};

static inline void outd(int16_t port, int32_t dword)
{
	asm volatile(
		  "outl %0, %1"
		:
		: "a"(dword), "Nd" (port)
	);

	return;
};

static inline uint8_t inb(int16_t port)
{
	register uint8_t byte;
	
	asm volatile(
		  "inb %1, %0"
		: "=a" (byte)
		: "Nd" (port)
	);
	
	return byte;
};

static inline uint16_t inw(int16_t port)
{
	register uint16_t word;
	
	asm volatile(
		  "inw %1, %0"
		: "=a" (word)
		: "Nd" (port)
	);
	
	return word;
};

static inline uint32_t ind(int16_t port)
{
	register uint32_t dword;
	
	asm volatile(
		  "inl %1, %0"
		: "=a" (dword)
		: "Nd" (port)
	);
	
	return dword;
};

static inline void insw(unsigned short port, unsigned int count, unsigned short *buf)
{
	asm volatile (
		"rep ; insw"
		: "=D" (buf), "=c" (count)
		: "d" (port), "0" (buf), "1" (count)
	);
	
	return;
};

static inline void cli(void)
{
	asm volatile(
		"cli"
	);
	
	return;
};

static inline void sti(void)
{
	asm volatile(
		"sti"
	);
	
	return;
};

#endif /* !___I386_PORTIO_H_ */

