/******************************************************************
 *
 * syscall.S
 * System call interface for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ******************************************************************/

.intel_syntax noprefix
.text
.code32

.extern idt_addblock_stdi
.extern idt_update
.extern CpuCheckFlag
.extern KsSyscallTable
.globl KsSyscall
KsSyscall:
	push [syscall_int]		# get isr entry
	push 0x50				# int no
	call idt_addblock_stdi	# add it!
	
	add esp,8				# don't need those on the stack
	
	call idt_update			# have to update the IDT...

	push 0x00000800			# SEP flag
	call CpuCheckFlag		# does this CPU support SEP?
	add esp,4				# pop the flag off the stack
	cmp eax,0				# check...
	je .setupTable			# if not, don't do MSR crap
	
	xor edx,edx				# clear top 32-bits
	mov eax,0x08			# value of CS in ring 0
	mov ecx,0x174			# IA32_SYSENTER_CS
	wrmsr					# write it!
	
	call __KiGetStack		# get our stack
	mov ecx,0x175			# IA32_SYSENTER_ESP
	wrmsr					# write it!
	
	lea eax,syscall_fast	# get EIP
	mov ecx,0x176			# IA32_SYSENTER_EIP
	wrmsr					# write it!
	
.setupTable:
	call KsSyscallTable		# setup table
	
	ret						# done

.extern _KiTranslateIsrSyscall
.globl syscall_int

syscall_int:
	push 0x0
	push 0x50
	call _KiTranslateIsrSyscall
	iret

# IMO the above is rather kludgey.  Like, Windows 95-type kludgey.
# Using an ISR declaration for syscalls is going to be *VERY SLOW*!
# This is much cleaner, and reminiscent of everyone's favourite
# arches (Jacob's MIPS, Daniel's PPC). It's also something modern,
# which supposedly AWOS doesn't have much of.
# According to an application note I just read, modern NT-based
# kernels use this kind of method too, only they actually set
# edx and ecx in user mode.  We don't want that, because edx and
# ecx can be used as parameters with fastcall.
# YES, WE"RE USING __fastcall IN KERNEL MODE.  DEAL WITH IT.
# see the wiki for the fast-syscall calling convention.

.extern _KiDispatchSyscall
.globl syscall_fast
.globl _KiRealDispatchSyscall

.align 4
syscall_fast:
	push eax		# save the EIP, because gcc will step on it
	push ebx		# user stack
	push ebx		# again, as function argument (fastcall only uses ecx and edx)
	call _KiDispatchSyscall
	pop ecx			# restore target stack pointer to ECX, where SYSEXIT expects 
	pop edx			# restore target instruction to EDX, where SYSEXIT expects
	sysexit			# and we'll see you next time.

_KiRealDispatchSyscall:
	# this is where the magic happens -- hax so leet, C can't even pull it off
	# basically, we have the fnptr in ecx, and we have the user's stack in edx
	# so all we have to do is literally mov esp,edx and call ecx
	# et voila, the function is called, in kernel mode, with the user stack.
	push ebp
	mov ebp,esp		# SAVE our stack, else Bad Things happen
	mov esp,edx		# load the stack
	call ecx		# call the function
	mov esp,ebp		# restore our stack
	pop ebp
	ret				# done
	# see how simple that is!

# Simple lil usermode test
# If all goes well, will actually print 1 on the screen, then exit.
.globl usermode
usermode:
	push '1'	# give the character "1" to it
	mov ebx,esp	# save our stack
	mov ecx,1	# putch()
	xor edx,edx	# don't need return value
	lea eax,$+8	# our EIP to return to
	sysenter	# call the system!!!!
	add esp,4	# pop the 1
	push 0		# successful exit code
	mov ebx,esp	# our stack
	mov ecx,0	# exit()
	xor edx,edx	# won't even be returning
	lea eax,$+8	# our EIP to return to (not that we will)
	sysenter	# call exit, we're done
	ret 1		# shouldn't ever see this -- we're dead after sysenter
