#################################################################
#
#kernel/v86.asm
#V86 entry/exit routines
#
#See LICENSE file included with this source for copyright
#information.
#
################################################################# 

.intel_syntax noprefix
.text
.code32

.global KiV86Enter
.global __KiV86Return_main

#.extern currtask
.extern ktss

.extern __KiUpdateTSS
.extern printf

#v86ent_rant:
#	.ascii "v86: seg=%04X ip=%04X sp=%04X",10,0

__KiV86Enter_main:
	pusha
	push ds
	push es
	push fs
	push gs
	# OK, there's our 32-bit stack frame.
	# We will now proceed with our 16-bit jumper.
	xor eax,eax
	xor ecx,ecx
	xor edx,edx
	mov ax,[ebx+0x08] # seg
	mov cx,[ebx+0x0C] # ip
	mov dx,[ebx+0x10] # sp
	
	## Debug info: I hate these parameter issues.
	# Actually, I'll give this a miss for now.
	# I don't need it, and it probably does more harm than good.
	#pusha
	#push edx
	#push ecx
	#push eax
	#push v86ent_rant
	#call printf
	#add esp,0x10
	#popa
	
	.rept 5 
		push eax # virtual {d/e/f/g}s, and user ss
	.endr
	
	# Sort the stack out.
	sub dx,4
	push edx # user sp
	
	# Add the IRET return code.
	mov edx,ecx # Yay, we can now abuse edx!
	shl edx,4
	add edx,[esp]
	mov dword ptr [edx], 0xFEEB4C21
	
	pushf
	mov edx,0x20200 # VM/IF
	or [esp],edx
	
	push eax # cs
	push ecx # ip
	sub esp,8 # Alignment hack, AKA "Let's do it like we've always done it" hack.
	
	# Update TSS {esp/ss}{0/1/2}.
	# We'll use the API this time.
	pusha
	mov eax,esp
	sub eax,0x10
	push eax
	call __KiUpdateTSS
	pop eax
	popa
	
	add esp,8
	
	#mov ax,ss
	#mov [ktss+0x04],ebx
	#mov [ktss+0x08],eax
	#mov [ktss+0x0C],ebx
	#mov [ktss+0x10],eax
	#mov [ktss+0x14],ebx
	#mov [ktss+0x18],eax
	
	# Blank it out.
	xor eax,eax
	xor ebx,ebx
	xor ecx,ecx
	xor edx,edx
	xor esi,esi
	xor edi,edi
	xor ebp,ebp
	
	# Now for the magic IRET.
	iret

__KiV86Return_main:
	add esp,(4+2)*4 # skip V86-specific data
	#int 0x00 # A zero division is a good way to shout out your results.
	pop gs
	pop fs
	pop es
	pop ds
	popa
	iret

KiV86Enter:
	push ebx
	
	mov ebx,esp
	
	pushf
	push cs
	call __KiV86Enter_main
	
	pop ebx
	ret
