/***************************************************************
 *
 * drivers/ata/ata.h
 * defines and structs for AWOS ATA driver
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#ifndef ___ATA_DRIVER_ATA_H_
#define ___ATA_DRIVER_ATA_H_

#include <kernel.h>
#include <cpuspec.h>

struct ata_disk_ident
{
	uint16_t Conf;
	uint16_t NumCylinders;
	uint16_t reserved_1;
	uint16_t NumHeads;
	uint16_t VendSpec_1;
	uint16_t VendSpec_2;
	uint16_t SectPerTrk;
	uint16_t VendSpec_3[3];
	char SerialNo[20];
	uint16_t VendSpec_4;
	uint16_t VendSpec_5;
	uint16_t LongSize;
	char FirmwareRev[8];
	char ModelName[40];
	uint16_t MaxSectPerBlk;
	uint16_t reserved_2;
	uint16_t Caps;
	uint16_t reserved_3;
	uint16_t PioTimingMode;
	uint16_t Obsolete;
	uint16_t RestOfStructValid;
	/** These fields ONLY VALID if bit 0 of RestOfStructValid is SET **/
	uint16_t NumLogCylinders;
	uint16_t NumLogHeads;
	uint16_t NumLogSectPerLogTrk;
	uint32_t CapacityInSectors;
	/** These fields are supposedly always valid **/
	uint16_t MultipleSectorSetting;
	uint32_t Capacity;
	uint16_t Obsolete2;
	uint16_t MultiwordDma;
	/** These fields ONLY VALID if bit 1 of RestOfStructValid is SET **/
	uint16_t PioModesSupported;
	uint16_t MinTimeMwDma;
	uint16_t RecTimeMwDma;
	uint16_t MinTimePio;
	uint16_t MinTimePioIordy;
	uint16_t reserved_4[10];
	uint16_t MajorVersion;	/** MSB# set = ATA ver, i.e. bit3=1 = ATA-3, bit4=1 = ATA4 */
	uint16_t MinorVersion;
	uint32_t CommandSetSupp;
	uint16_t reserved_5[43];
	uint16_t SecStatus;
	uint16_t VendSpec_6[30];
	uint16_t reserved_6[95];
};

/* gah, won't compile with this. 
const char *MinorVersion[] = {
	"ATA (ATA-1) prior to Revision 4", "ATA-1", "ATA-1 revision 4", "ATA-2",
	"ATA-2 prior to Revision 2k", "ATA-3 Revision 1", "ATA-2 Revision 2k", "ATA-3 Revision 0",
	"ATA-2 revision 3", "ATA-3", "ATA-3 Revision 6", "ATA-3 Revision 7"
};
*/

#define ATA_CAP_TIMER	0x1000
#define ATA_CAP_IORDY	0x0400
#define ATA_CAP_IORDYD	0x0200

typedef struct ata_disk_ident * ata_disk_ident_t;

typedef struct ata_disk * AtaDisk;

typedef struct __atacontroller
	{
		/** base port (ex 0x1F0) **/
		uint16_t	basePort;
		/** IRQ **/
		uint8_t		irq;
		
		AtaDisk		Disks[2];
		
		/** Set to TRUE when waiting for an IRQ -- will be cleared when IRQ done **/
		BOOL		 ExpectingIRQ;
	} AtaController;

struct ata_disk
{
	/** Controller parent **/
	AtaController *parent;
	/** Type of device structure **/
	uint8_t Type;
	/** disk number on controller (0 or 1) **/
	uint8_t DiskNumber;
	/** is the disk present on the controller? **/
	BOOL Present;
	/** Defines whether this disk is in a lock/mutex
	    XXX: when we have a working getpid(), should
	    be set to that value. **/
	uint32_t Locked;
	/** identification information **/
	ata_disk_ident_t Ident;
};

enum AtaPowerModes
{
	AtaPowerMode_Active = 0x00,
	AtaPowerMode_Idle = 0x80,
	AtaPowerMode_IdleImmediate,
	AtaPowerMode_Sleep,
	AtaPowerMode_StandbyImmediate,
	AtaPowerMode_Standby = 0xFF
};

BOOL AtaIsValidController(uint16_t);

AtaController *AtaControllerInit(uint8_t, uint16_t);
void AtaControllerDestroy(AtaController *);

/* driver stuff */
void AtaIrqHandler(void *);

/* ATA stuff */
AtaDisk		 AtaObtainDisk(AtaController *, uint8_t);
void		 AtaReleaseDisk(AtaDisk);
char		*AtaRead(AtaDisk, BOOL, uint8_t, uint32_t);
BOOL		 AtaWrite(AtaDisk, BOOL, uint8_t, uint32_t, char *);
uint8_t		 AtaGetPowerMode(AtaDisk);
BOOL		 AtaSetPowerMode(AtaDisk, uint8_t);
BOOL 		 AtaSetTransferMode(AtaDisk, uint8_t, uint8_t);
BOOL		 AtaSetDoorLock(AtaDisk, BOOL);
BOOL		 AtaEject(AtaDisk);

uint8_t		 AtaReadRegisterBsyWait(AtaController *, uint8_t);
void		 AtaWriteRegister(AtaController *, uint8_t, uint8_t);
char		*AtaGetSector(AtaController *, uint8_t);
ata_disk_ident_t AtaGetIdent(AtaDisk Disk);

#define DRVTYPE_DISK_ATA	0x01

#define ATA_STATUS_ERR	0x01	/* error */
#define ATA_STATUS_IDX	0x02	/* index -- vendor specific */
#define ATA_STATUS_CORR	0x04	/* corrected data */
#define ATA_STATUS_DRQ	0x08	/* data request */
#define ATA_STATUS_DSC	0x10	/* device seek complete */
#define ATA_STATUS_DF	0x20	/* device fault */
#define ATA_STATUS_DRDY	0x40	/* device ready */
#define ATA_STATUS_BSY	0x80	/* busy */

#define ATA_ERROR_AMNF	0x01	/* Address Mark Not Found */
#define ATA_ERROR_TK0NF	0x02	/* Track 0 Not Found */
#define ATA_ERROR_ABRT	0x04	/* Aborted Command */
#define ATA_ERROR_MCR	0x08	/* Media Change Requested */
#define ATA_ERROR_IDNF	0x10	/* ID Not Found */
#define ATA_ERROR_MC	0x20	/* Media Changed */
#define ATA_ERROR_UNC	0x40	/* Unrecoverable Data Error */
#define ATA_ERROR_RES	0x80	/* Reserved as of ATA-3 */

#define ATA_REG_DATA		0x0	/* READ/WRITE */	/* Read only when status & DRQ = 1 */

#define ATA_REG_ERROR		0x1	/*  READ ONLY */	/* valid when BSY=0 and DRQ=0 and ERR=1 */
/* this register is sometimes used for command parameters */
#define ATA_REG_FEATURES	0x1	/* WRITE ONLY */	/* valid when BSY=0 and DRQ=0 */

/* if the contents of the register are 0, 256 is assumed (i.e. purposeful overflow) */
#define ATA_REG_SECCOUNT	0x2	/* READ/WRITE */	/* valid when BSY=0 and DRQ=0 */

#define ATA_REG_SECNUM		0x3	/* READ/WRITE */	/* valid when BSY=0 and DRQ=0 -- sec7:0 or lba7:0*/
#define ATA_REG_LBALOW		0x3	/* READ/WRITE */

/* Write to these registers only after checking that status & BSY = 0 and status & DRQ = 0 */
#define ATA_REG_CYL_LO		0x4	/* READ/WRITE */	/* cylinder  7:0 or LBA 15:8  */
#define ATA_REG_LBAMID		0x4 /* READ/WRITE */
#define ATA_REG_CYL_HI		0x5	/* READ/WRITE */	/* cylinder 15:8 or LBA 23:16 */
#define ATA_REG_LBAHIGH		0x5	/* READ/WRITE */
#define ATA_REG_DEVHEAD		0x6	/* READ/WRITE */	/* b7 = 1, b6 = LBA, b5 = 1, b4 = DRV, b3-b0 = LBA 27:24 */
/*#define ATA_REG_DEVCTL		0x6	 WRITE ONLY */
/*#define ATA_REG_STATUS_ALT	0x6	  READ ONLY */

#define ATA_REG_STATUS		0x7	/*  READ ONLY */	/* if BSY=1 in this reg, all other flags are invalid */
#define ATA_REG_COMMAND		0x7	/* WRITE ONLY */

/*char *AtaRead(void *AtaDisk, BOOL Retry, uint8_t NumSectors, uint32_t StartSector);
BOOL AtaWrite(void *AtaDisk, BOOL Retry, uint8_t NumSectors, uint32_t StartSector, char *data);

uint8_t AtaGetPowerMode(void *AtaDisk);
BOOL AtaSetPowerMode(void *AtaDisk, uint8_t Mode);

BOOL AtaSetDoorLock(void *AtaDisk, BOOL Lock);

ata_disk_ident_t AtaGetIdent(ata_disk_t);

BOOL AtaEject(void *AtaDisk);

BOOL AtaSetTransferMode(void *AtaDisk, uint8_t Type, uint8_t Mode);

BOOL AtaObtainLock(void *);
void AtaReleaseLock(void *);

struct DrvInfo_Ata *AtaInit(uint16_t port, uint8_t irq);*/

#endif /* !___ATA_DRIVER_ATA_H_ */

