/***************************************************************
 * 
 * drivers/comm/uart_io.c
 * UART input/output routines for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <portio.h>
#include <stdint.h>
#include "uart.h"

//void csleep(uint16_t);
void yield(void);

void SerSetUartParams(int16_t port, int16_t bps, uint8_t flags)
{
	outb(port + SER_REG_LINECTL, 0x80);	/* set registers to DIVISOR */
	outw(port + SER_REG_DATA, bps);		/* set baud rate */
	outb(port + SER_REG_LINECTL, flags);	/* set misc options */
	outb(port + SER_REG_MODEMCTL, 0x0);	/* no loopback */
	
	return;
};

void SerSendChar(int16_t port, char c)
{
	while(((inb(port + SER_REG_LINESTAT)) & SER_LSR_THRE) == 0) yield();
	outb(port + SER_REG_DATA, c);

	return;
};

char SerDetectUart(int16_t port)
{
	int16_t oldModemCtl, oldScratch, fifoStatus;
	
	oldModemCtl = inb(port + SER_REG_MODEMCTL);	// save original contents
	outb(port + SER_REG_MODEMCTL, 0x10);		// loopback mode
	
	if((inb(port + SER_REG_MODEMSTAT) & 0xF0) == 0) return 0;		// no UART
	
	outb(port + SER_REG_MODEMCTL, 0x1F);		// toggle flags
	
	if((inb(port + SER_REG_MODEMSTAT) & 0xF0) != 0xF0) return 0;	// no UART
	
	/** UART is now known to exist.  all we need is type **/
	outb(port + SER_REG_MODEMCTL, oldModemCtl);	// restore the original contents
	
	oldScratch = inb(port + SER_REG_SCRATCH);	// save original scratch
	outb(port + SER_REG_SCRATCH, 0x55);			// are we up to scratch?
	
	if(inb(port + SER_REG_SCRATCH) != 0x55) return 1;	// we are not.
	
	outb(port + SER_REG_SCRATCH, oldScratch);	// restore original
	
	outb(port + 0x2, 0x01);						// enable FIFO
	fifoStatus = inb(port + 0x2);				// get status
	outb(port + 0x2, 0x00);						// disable FIFO
	
	if((fifoStatus & 0xC0) == 0) return 2;		// 8250A/16450
	if((fifoStatus & 0xC0) == 0x80) return 3;	// 16550 (no FIFO)
	if((fifoStatus & 0xC0) == 0xC0) return 4;	// 16550A
	
	return 0;									// no clue.
};
