/***************************************************************
 *
 * drivers/pci/usb.c
 * USB driver for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

// Maybe I'm trying a little *too* hard here.
// Maybe that's what PCI coding does to you.
// Well, I've never done this before, so this'll be interesting...
//    -GM

// Right, seeing as USB is rather rare in a hobby OS,
// I'm going to comment the hell out of this. Enjoy. -GM

#include <kernel.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <portio.h>	/* inX, outX */

#include <pci.h>
#include <usb.h>

// Just because there's no particularly standard scheduler interface yet...
#include <sched.h>

// This is BAD BAD BAD BAD
// We can't use this driver on other arches because of this.
// TODO XXX BUG HACK !!!!
#include <cpuspec.h>

BOOL usb_irq_status_stfu = TRUE; // Supress IRQ status messages?

struct usbhost_ll *usbhub_irq_host[16]; // Just to tell which IRQ is for which hub.
struct usbhost_ll *usbroot = NULL, *usbrootend = NULL;

// UsbUhciPortIn8/16/32 and UsbUhciPortOut8/16/32 are for generic input/output.
void UsbUhciPortOut8(struct usbhost_ll *host, uint16_t usbport, uint8_t data)
{
	outb((int16_t)(host->port | (usbport & host->portmask)), data);
}

void UsbUhciPortOut16(struct usbhost_ll *host, uint16_t usbport, uint16_t data)
{
	outw((int16_t)(host->port | (usbport & host->portmask)), data);
}

void UsbUhciPortOut32(struct usbhost_ll *host, uint16_t usbport, uint32_t data)
{
	outd((int16_t)(host->port | (usbport & host->portmask)), data);
}

uint8_t UsbUhciPortIn8(struct usbhost_ll *host, uint16_t usbport)
{
	return inb((int16_t)(host->port | (usbport & host->portmask)));
}

uint16_t UsbUhciPortIn16(struct usbhost_ll *host, uint16_t usbport)
{
	return inw((int16_t)(host->port | (usbport & host->portmask)));
}

uint32_t UsbUhciPortIn32(struct usbhost_ll *host, uint16_t usbport)
{
	return ind((int16_t)(host->port | (usbport & host->portmask)));
}


// UsbUhciCmdSet and UsbUhciCmdClear set and clear UHCI command flags.
void UsbUhciCmdSet(struct usbhost_ll *host, uint16_t flags)
{
	UsbUhciPortOut16(host, USB_UHCI_USBCMD, (int16_t)(UsbUhciPortIn16(host, USB_UHCI_USBCMD) & flags));
}

void UsbUhciCmdClear(struct usbhost_ll *host, uint16_t flags)
{
	UsbUhciPortOut16(host, USB_UHCI_USBCMD, (int16_t)(UsbUhciPortIn16(host, USB_UHCI_USBCMD) & ~flags));
}

// UsbUhciStatusRead and UsbUhciStatusClear read and clear UHCI status flags.
uint16_t UsbUhciStatusRead(struct usbhost_ll *host)
{
	return UsbUhciPortIn16(host, USB_UHCI_USBSTS);
}

void UsbUhciStatusClear(struct usbhost_ll *host, uint16_t flags)
{
	UsbUhciPortOut32(host, USB_UHCI_USBSTS, flags); // sending a 1 clears the selected flag according to the UHCI spec.
}

// UsbUhciFrameAddrUpdate updates the frame list by the usbhost_ll structure given.
void UsbUhciFrameAddrUpdate(struct usbhost_ll *host)
{
	int cmd;
	
	// We do not want to be interrupted during this transaction.
	cli();
	cmd = UsbUhciPortIn16(host, USB_UHCI_USBCMD);
	if(cmd & USB_UHCI_CMD_RS)
		UsbUhciCmdClear(host, USB_UHCI_CMD_RS);
	
	host->frameidx &= 0x7FF; // Sanity check + update
	UsbUhciPortOut32(host, USB_UHCI_FLBASEADD, (uint32_t)host->framelist);
	UsbUhciPortOut16(host, USB_UHCI_FRNUM, 0);
	
	// OK, all clear.
	if(cmd & USB_UHCI_CMD_RS)
		UsbUhciCmdSet(host, USB_UHCI_CMD_RS);
	sti();
}

// UsbUhciCreateTD builds a basic transfer descriptor.
struct usb_uhci_td_t *UsbUhciCreateTD(void *buffer, uint16_t maxlen, uint8_t pid)
{
	struct usb_uhci_td_t *td;
	void *v;
	
	v = malloc(sizeof(struct usb_uhci_td_t) + 0xF); // This must be aligned to 16.
	if(v == NULL) // We must allocate it!
		return NULL;
	
	td = (void *)(((uint32_t)v+0xF) & ~0xF); // Align.
	td->actualbuf = v;
	
	td->linkptr = USB_UHCI_TDLP_TERM; // No link pointer for now.
	
	// TODO: find out what we need for this!
	// Feel free to experiment until something works.
	td->status = 0;
	td->status |= USB_UHCI_TDCS_SACTIVE;
	td->status |= USB_UHCI_TDCS_INTONCPL;
	td->status |= USB_UHCI_TDCS_ISOCRON;
	
	// Alter maxlen to suit.
	if (maxlen == 0)
		maxlen = 0x7FF; // This is the equivalent of zero.
	else if (maxlen > 0x500)
		maxlen = 0x4FF; // This is as high as we can go.
	else
		maxlen--; // maxlen has to be adjusted: 0x000 = 1 byte.
	
	// Insert the maximum length.
	td->token = (maxlen << USB_UHCI_TDTK_SHLMAXLEN) & USB_UHCI_TDTK_MSKMAXLEN;
	td->token |= pid & USB_UHCI_TDTK_MSKPKTID;
	td->buffptr = (uint32_t)buffer;
	
	return td;
}

// UsbUhciDestroyTD frees a transfer descriptor.
void UsbUhciDestroyTD(struct usb_uhci_td_t *td)
{
	free(td->actualbuf);
}

// UsbUhciScheduleTD attempts to schedule the transfer descriptor.
BOOL UsbUhciScheduleTD(struct usbhost_ll *host, struct usb_uhci_td_t *td)
{
	// FIXME: this should really check to see if there's something there first.
	
	cli(); // Clear interrupts.
	
	host->frameidx &= 0x3FF; // SANITY: don't accept any substitutes.
	host->framelist[host->frameidx] = ((uint32_t)td & ~0xF) | USB_UHCI_FLP_TD; // Inject.
	host->frameidx++; // Increment frame pointer.
	
	sti(); // OK, all clear now.
	
	return TRUE;
}

// irq_handler_usb() is, well, self-explanitory.
void irq_handler_usb(void *irqparams)
{
	struct usbhost_ll *h;
	int status;
	int frame;
	struct regs *r = (struct regs *)irqparams;
	
	// Check which hub we want.
	h = usbhub_irq_host[r->int_no];
	if(h == NULL)
		return; // We only want IRQs with hubs assigned!
	
	// Get the status, then shut it up for now.
	status = UsbUhciStatusRead(h);
	//UsbUhciCmdClear(h, USB_UHCI_CMD_RS);
	
	// Check the frame.
	frame = UsbUhciPortIn16(h, USB_UHCI_FRNUM);
	
	// Spit out the status + frame details, unless usb_irq_status_stfu is set.
	if(usb_irq_status_stfu == FALSE)
	{
		dprintf("usbd-irq: IRQ %i gives status %08X - frame %08X/%04X=%08X\n"
			,r->int_no
			,status
			,UsbUhciPortIn32(h, USB_UHCI_FLBASEADD)
			,frame
			,h->framelist[frame & 0x3FF] // Mask it.
				);
	};
	
	// Clear the status.
	UsbUhciStatusClear(h, 0xFFFF);
	
	// That's all we have so far.
	//UsbUhciCmdSet(h, USB_UHCI_CMD_RS);
}

// daemon_usbd() is a USB scheduler task.
void daemon_usbd(void)
{
	struct usbhost_ll *h;
	struct usb_uhci_td_t *td;
	int i;
	
	// (void)td;
	
	dprintf("usbd: Starting USB daemon\n");
	
	for(;;)
	{
		// Check every hub.
		for(h = usbroot ; h != NULL ; h = h->n)
		{
			// Is this hub configured? If not, then configure it!
			if(!(h->flags & USB_AWOS_CONFIGURED))
			{
				dprintf("usbd: Setting up host %04X:%i:%i:%i at %04X\n"
					,h->parent->port
					,h->parent->b
					,h->parent->s
					,h->parent->f
					,h->port
					);
				
				// Set up IRQ hook.
				i = (h->parent->dev).irq_line;
				dprintf("usbd: Hooking IRQ %i\n", i);
				usbhub_irq_host[i] = h;
				KiIrqInstall(i, irq_handler_usb, "USB handler");
				
				// Reset the hub.
				UsbUhciCmdSet(h, USB_UHCI_CMD_HCRESET);
				while(UsbUhciPortIn16(h, USB_UHCI_USBCMD) & USB_UHCI_CMD_HCRESET)
					yield();
				dprintf("usbd: Host reset\n");
				
				// Assign a frame list, align to 4KB, and set count to 1.
				// TODO: malloc_align() function.
				h->framelist_malloc = malloc(4096 + 4095);
				h->framelist = (void *)(( (uint32_t)(h->framelist_malloc) + 0xFFF ) & ~0xFFF);
				h->frameidx = 0;
				
				// Set terminate flag on all items in list frame pointer table.
				for(i=0; i<1024; i++)
					h->framelist[i] = USB_UHCI_FLP_TERM;
				
				// Initialise index.
				h->frameidx = 0;
				
				// Update the frame pointer / index.
				UsbUhciFrameAddrUpdate(h);
				
				// Start it up.
				UsbUhciCmdSet(h, USB_UHCI_CMD_CF);
				while(UsbUhciPortIn16(h, USB_UHCI_USBCMD) & USB_UHCI_CMD_CF)
					yield();
				dprintf("usbd: Configured\n");
				
				UsbUhciCmdSet(h, USB_UHCI_CMD_RS);
				//while(!(UsbUhciPortIn16(h, USB_UHCI_USBCMD) & USB_UHCI_CMD_RS))
				//{
				//	UsbUhciCmdSet(h, USB_UHCI_CMD_RS);
				//	yield();
				//}
				
				dprintf("usbd: Ready\n");
				
				// Yay. *NOW* we're configured.
				h->flags |= USB_AWOS_CONFIGURED;
			} else {
				// Is it halted?
				if(UsbUhciStatusRead(h) & USB_UHCI_STS_HCHLT)
				{
					// Is a frame waiting to execute?
					i = UsbUhciPortIn16(h, USB_UHCI_FRNUM) & 0x3FF;
					if(!(h->framelist[i] & USB_UHCI_FLP_TERM))
					{
						// Check the transfer info.
						td = (void *)(h->framelist[i] & ~0xF);
						
						// Is this an active transfer?
						if(td->status & USB_UHCI_TDCS_SACTIVE)
						{
							// Clear the status.
							UsbUhciStatusClear(h, 0xFFFF);
							
							// Start the hub.
							UsbUhciCmdSet(h, USB_UHCI_CMD_RS);
						} else {
							// Terminate the transfer.
							h->framelist[i] |= USB_UHCI_FLP_TERM;
							dprintf("usbd: Transfer %03X terminated with status %08X\n", i, td->status);
							// Test something...
							dprintf("Result: (%08X)%08X=%04X\n", td, td->buffptr, *((uint16_t *)(td->buffptr)));
						};
					};
				};
			};
		};
		
		// Let's not rape the CPU, shall we not?
		yield();
	};
	
	// Just in case.
	// XXX - ICC says this is unreachable, so i'm commenting it out -- Jacob
	// dprintf("usbd: Exiting... why?!\n");
	// exit(0);
}

// UsbAddPciHost is used to grab a USB bus from a PCI->USB bridge.
void UsbAddPciHost(struct pcidev_ll *host)
{
	struct usbhost_ll *newhost;
	int m,r;
	
	// SANITY: We don't want to have a null host by mistake.
	if(host == NULL)
		return;
	
	if((host->dev).bar[4] & 1) // UHCI spec explicitly states use of BAR4. *WE ONLY WANT I/O PORTS!*
	{
		// We *MUST* ensure that this is sane before we insert it.
		newhost = malloc(sizeof(struct usbhost_ll));
		
		// Clear the data first
		newhost->port = newhost->portmask = 0x0000;
		newhost->n = NULL;
		// Right, we're going to determine the size of BAR4
		// by changing the value.
		r = (host->dev).bar[4];
		newhost->port = (int16_t) (r & ~3);
		
		cli(); // We're screwing things up here; we don't want an interrupt to bug us!
		
		PciWriteConf32((int16_t)host->port, host->b, host->s, host->f, 0x20, 0x0000FFFF); // Fill with 1s
		m = PciReadConf32((int16_t)host->port, host->b, host->s, host->f, 0x20); // Get the mask (we invert it later)
		PciWriteConf32((int16_t)host->port, host->b, host->s, host->f, 0x20, r); // Now return BAR4 to its old value.
		
		sti(); // OK, now it's unscrewed.
		
		newhost->portmask = (int16_t)((m ^ 0xFFFF) | 3); // First bit is always 1 for a port request.
		
		newhost->parent = host; // Backlink for PCI info.
		
		dprintf("usb: assigned host at ports %04X-%04X", newhost->port, newhost->port|newhost->portmask);
		r = PciReadConf32((int16_t)host->port, host->b, host->s, host->f, 0x60) & 0x000000FF;
		switch(r) // This is what the UHCI spec tells me about version info...
		{
			case 0x00:
				dprintf(" (pre-1.0 UHCI hub)\n");
				break;
			case 0x10:
				dprintf(" (1.0 UHCI hub)\n");
				break;
			default:
				dprintf(" (unknown ver %02X)\n", (unsigned)r);
		};
		
		// Right, *NOW* we inject it into the list.
		
		cli(); // Be careful here.
		
		// If usbroot == NULL, create it, else tag it in.
		if(usbrootend == NULL)
		{
			usbrootend = usbroot = newhost;
		} else {
			// NOTE: this code can be a bit confusing, but it shouldn't screw up.
			// Think of it like this without the "usbrootend =" bit.
			usbrootend = usbrootend->n = newhost;
		};
		
		sti(); // Now we return to carelessness.
	};
}

// UsbInit is called after PciInit to initialise the usb driver.
void UsbInit(void)
{
	struct pcidev_ll *host = NULL;
	int i;
	
	printf("usb: Building USB host database\n");
	
	if(k_kcmdline_check("usbnoisy") == NULL) // Suppress status messages?
	{
		usb_irq_status_stfu = TRUE;
		dprintf("usb: Suppressing excess status messages.\n");
	} else {
		usb_irq_status_stfu = FALSE;
		dprintf("usb: Allowing excess status messages. THESE CAN BE NOISY!\n");
	};
	
	// i is used to give a USB hub count.
	// We will find USB hosts via PCI.
	for(i=0 ; ; i++)
	{
		// This is the PCI-to-USB bridge class. Find all of these devices.
		host = PciFindByClassB(host, PCI_CLASS_SERIAL, PCI_CLASS_SERIAL_USB);
		if(host == NULL) // NULL means there are no more PCI devices of this class.
			break;
		
		dprintf("usb: found host %04X:%04X at %04X:%i:%i:%i\n"
			,(host->dev).vendor_id
			,(host->dev).device_id
			,host->port
			,host->b
			,host->s
			,host->f
				);
		
		UsbAddPciHost(host);
	};
	if(i == 1)
		printf("usb: Found 1 USB host via PCI\n");
	else
		printf("usb: Found %i USB hosts via PCI\n", i);
	
	// Load usbd.
	__KiCreateTask((uint32_t)daemon_usbd, "usbd");
}
