/****************************************************************

kernel/main.c
Entry point for AWOS Kernel

See LICENSE file included with this source for copyright
information.

****************************************************************/
extern "C"
{
#include <kernel.h>
#include <cons.h>
#include <input.h>
#include <kalloc.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <portio.h>
#include <sched.h>
#include <pci.h>
#include <usb.h>
#include <fs.h>
#include <dma.h>
#include <v86.h>
#include <cpuinfo.h>

#include "drivers/ata/ata.h"
#include "net/ifaces/e1000.h"
#include "net/ifaces/ne2000.h"
#include "net/ifaces/rtl8139.h"
#include "net/hal/iface-mgr.h"

void init(void);
#include "drivers/comm/uart.h"
extern int16_t DebugPort;
AtaController *pController = NULL;
AtaController *sController = NULL;
Net_E1000IfDrv *e1000 = NULL;
Net_NE2000IfDrv *ne2000 = NULL;
Net_Rtl8139IfDrv *rtl8139 = NULL;
extern void KsSyscall(void);
extern void Init440(void);
extern void InitNvidia(void);
struct KiCpuInfo *cpuInfo;

BOOL debug = FALSE;

void _MmTest();
}; /* extern "C" */

extern void shell(void);

void vga_sexypattern(void)
{
	int x, y, i, j;
	
	for(y=0; y<200; y++)
	{
		j = y/25;
		
		currtty->driver->gseek(currtty, 0,y);
		for(x=0; x<320; x+=10,j+=8)
		{
			for(i=0; i<10; i++)
			{
				currtty->driver->plot(currtty, j);
				currtty->driver->gincx(currtty);
			};
		};
	};
}

void vesa_sexypattern(void)
{
	int x, y, i, j;
	
	for(y=0; y<480; y++)
	{
		currtty->driver->gseek(currtty, 0,y);
		for(x=0; x<640; x+=10,j+=8)
		{
			for(i=0; i<10; i++)
			{
				currtty->driver->plot(currtty, x + (y<<8));
				currtty->driver->gincx(currtty);
			};
		};
	};
}

multiboot_info_t *mbinfo;

char *k_cmdline_check(char *cmdline, char *str)
{
	char *a;
	for(;;)
	{
		while((' '|*cmdline) != ' ')
			cmdline++;
		
		if(*cmdline == '\0')
			return NULL;
		
		cmdline++;
		
		a = str;
		while(*cmdline == *a && *a != '\0')
		{
			cmdline++;
			a++;
		};
		if(*cmdline == '=') {
			cmdline++;
			if(*a == '\0')
				return cmdline;
			
		} else if ((' '|*cmdline) == ' ') {
			if(*a == '\0')
				return cmdline;
		}
	};
}

char *k_kcmdline_check(char *str)
{
	return k_cmdline_check((char *)mbinfo->cmdline, str);
}

extern "C"
{
	void k_main(void *arch_boot, uint8_t arch)
	{
		int i;
		
		switch(arch)
		{
			case 0x00:	// x86
				mbinfo = (multiboot_info_t *)arch_boot;
				KsTableSetup();	// set up [g/i]dt
				break;
			case 0x01:	// PPC
				// Do some openfirmware crap
				for(;;) ;	// just loop
				break;
			default:
				printf("Unknown architecture %d", arch);
				panic("can't continue boot phase");
				break;
		};
		
		ConInit();
		
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"debug") != NULL)
			debug = TRUE;	
		
		printf("AWOS Version 0.1.00 [Build %s]\n", BUILD_DATE);
		printf("System memory: %lu/%lu kilobytes.\n", \
			mbinfo->mem_lower, mbinfo->mem_upper);
		printf("Kernel command line: %s\n", (char *)mbinfo->cmdline);
		if(debug == TRUE)
		{
			i = currtty->color;
			currtty->color = 0x0E;
			printf("\xB0\xB1\xB2\xDB DEBUG MODE \xDB\xB2\xB1\xB0\n");
			currtty->color = i;
		};
		
		/* Initialise memory manager */
		MmInit((void *)0x1FFFE0);
		MmAddToKernelHeap((void *)0x200000, 0x200000);
		MmAddToUserHeap((void *)0x400000, ((mbinfo->mem_upper * 1024) - 0x400000));
		
		/* Get CPU information -- should be platform independent... */
		cpuInfo = (struct KiCpuInfo *)kmalloc(sizeof(struct KiCpuInfo));
		KiCpuId(cpuInfo);
		
		/* you have to set up syscalls after CPU detection now. */
		KsSyscall();
		
		printf("Core set up, loading init...");
		__KiCreateTask((uint32_t)&init, (char *)"init");
		printf("done.\n");
		yield();
		return;
	}

	void init(void)
	{
		int i;
		uint16_t ttyS[4];
		char *c;
		console_t *tty;
	
		printf("INIT version 0.1 booting\n");
		
		/* v86 */
		KsV86Init();
		
		/* dma */
		DmaInit();
	
		/**** SERIAL TTY DETECTION CODE ****/
		ttyS[0] = 0x3F8;
		ttyS[1] = 0x2F8;
		ttyS[2] = 0x3E8;
		ttyS[3] = 0x2E8;
	
		for(i = 0; i < 3; i++)
		{
			char type;
			type = SerDetectUart(ttyS[i]);
			switch(type)
			{
				case 0: /* silently ignore existant ports */
					break;
				case 1: printf("ttyS%d: 8250-compatible\n", i);
					break;
				case 2:	printf("ttyS%d: 16450-compatible\n",i);
					break;
				case 3:	printf("ttyS%d: 16550-compatible\n",i);
					break;
				case 4:	printf("ttyS%d: 16550A-compatible\n",i);
					break;
			};
	
			if(type > 0)
			{
			SerSetUartParams(ttyS[i], SER_BPS_19200, SER_FLAGS_8N1);
			
			/* debug to lowest enumerated serial port, unlike NT */
			if(!DebugPort)
				if(k_kcmdline_check((char *)"ttydebug") == NULL)
					DebugPort = ttyS[i];
			};
		}
		/**** END SERIAL TTY DETECTION CODE ****/
		
		/* keyboard */
		KbdInit();
		
		/* PCI */
		PciInit();
		
		/* USB */
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"usb") != NULL)
		{
			printf("USB driver enabled, now initialising\n");
			UsbInit();
		};
		
		/* Filesystems */
		FsInit();
		
		/* vesa */
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"novesa") == NULL)
			VgaVesaInit();
		else
			printf("VESA detection skipped as requested on command line.\n");
		
		/* nvidia */
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"nonvidia") == NULL)
			InitNvidia();
		else
			printf("nVidia detection skipped as requested on command line.\n"); // Just in case!
		
		/* Graphics stuff */
		c = k_cmdline_check((char *)mbinfo->cmdline, (char *)"gfx");
		if(c != NULL)
		{
			if(!strcmpsp(c,(char *)"vga320")) {
				tty = VgaSetMode(320, 200, 8);
				if(tty)
				{
					VgaSetConsole(tty);
					VgaClearScreen();
					vga_sexypattern();
				} else {
					printf("Error: failed to obtain 320x200x8 graphics mode.\n");
				}
			} else if(!strcmpsp(c,(char *)"vesa640")) {
				tty = VgaSetMode(640, 480, 32);
				if(!tty)
					tty = VgaSetMode(640, 480, 24);
				
				if(tty)
				{
					VgaSetConsole(tty);
					VgaClearScreen();
					vesa_sexypattern();
				} else {
					printf("Error: failed to obtain 640x480x24/32 graphics mode.\n");
				}
			} else {
				printf("Mode not supported.\nSupported modes:"
					"- vga320\n"
					"- vesa640\n"
						);
			};
		};
		
		/* hdd */
		pController = AtaControllerInit(0x0E, 0x1F0);	//primary
		sController = AtaControllerInit(0x0F, 0x170);	//secondary

		/* 440BX */
		Init440();

		/* net */
		NetHalInterfaceStart();
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"e1000") != NULL)
			NetHalInterfaceAdd((Net_IfaceDriver *)new Net_E1000IfDrv());
		if(k_cmdline_check((char *) mbinfo->cmdline, (char *) "ne2000") != NULL)
			NetHalInterfaceAdd((Net_IfaceDriver *)new Net_NE2000IfDrv());
		if(k_cmdline_check((char *)mbinfo->cmdline, (char *)"rtl8139") != NULL)
			NetHalInterfaceAdd((Net_IfaceDriver *)new Net_Rtl8139IfDrv());
		
		printf("Time to rock.\n");
		
		/* start the shell */
		__KiCreateTask((uint32_t)&shell, (char *)"greasy-shell");
		
#ifndef __PPC
		while(1) asm("hlt");
#else
		while(1) ;
#endif
	}
}

