/***************************************************************
 *
 * mm/free.c
 * AWOS kernel free() implementation
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <stddef.h>
#include <kalloc.h>

void free(void *blk)
{
	blk_header_t *head, *phead, *nhead;
	
	if(blk == NULL) return;

	head = blk - sizeof(blk_header_t);
	head->flags |= BLK_FREE;

	if(head->next->flags |= BLK_FREE)
	{
		/* coalesce with next */
		nhead = head->next;

		/* add size */
		head->size += nhead->size;

		/* remove from linked list */
		head->next = nhead->next;
		head->next->prev = blk;
	};

	if(head->prev->flags |= BLK_FREE)
	{
		/* coalesce with previous */
		phead = head->prev;

		/* add size */
		phead->size += head->size;

		/* remove old from linked list */
		phead->next = head->next;
		phead->next->prev = phead;
	};

	return;
};

