/***************************************************************
 *
 * arch/i386/mm/include/kalloc.h
 * kernel allocator header
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#ifndef ___I386_MM_KALLOC_H_
#define ___I386_MM_KALLOC_H_

#include <stdint.h>
#include <sys/types.h>

/** memory block header **/
typedef struct blk_header
{
	/** previous header **/
	struct blk_header *prev;
	/** next header **/
	struct blk_header *next;
	/** is_blk_free, kernel vs user, etc **/
	uint8_t flags;
	/** size of block in bytes **/
	size_t size;
} blk_header_t;

/*** BLOCK HEADER FLAGS ***/
/** is block free? **/
#define BLK_FREE	0x01
/** if clear, it's kernel; if set, it's user **/
#define BLK_USER	0x02
/** if set, the block is paged into physical memory **/
#define BLK_PHYS	0x04

/** block list **/
extern blk_header_t *blk_list;

/*** FUNCTIONS ***/
void *kmalloc(size_t);
void kfree(void *);

void MmInit(void *);
void MmAddToKernelHeap(void *, size_t);
void MmAddToUserHeap(void *, size_t);

#endif /* !___I386_MM_KALLOC_H_ */

