/***************************************************************
 *
 * mm/mminit.c
 * AWOS memory management initialisation
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <kernel.h>
#include <kalloc.h>

blk_header_t *blk_list;

void MmInit(void *first_blk)
{
	dprintf("mm: Memory manager init, block list at %08p\n", first_blk);
	/* the block list begins. */
	blk_list = first_blk;
	blk_list->size = 0;
	blk_list->next = blk_list;
	blk_list->prev = blk_list;

	return;
};

void MmAddToKernelHeap(void *blk, size_t size)
{
	blk_header_t *head;

	dprintf("mm: block at %08p with size %d added to kernel heap\n", blk, size);

	head = blk;
	head->flags = BLK_FREE;
	/* add to linked list */
	head->prev = blk_list;
	head->next = blk_list->next;
	head->next->prev = head;
	head->prev->next = head;
	head->size = size;

	return;
};

void MmAddToUserHeap(void *blk, size_t size)
{
	blk_header_t *head;

	dprintf("mm: block at %08p with size %d added to user heap\n", blk, size);
	head = blk;
	head->flags = BLK_FREE | BLK_USER;
	/* add to linked list */
	head->prev = blk_list;
	head->next = blk_list->next;
	head->next->prev = head;
	head->prev->next = head;
	head->size = size;

	return;
};

