/***************************************************************
  *
  * net/hal/hal_driver.h
  * Hardware Abstraction Layer for the NIC drivers
  *
  * See LICENSE file included with this source for copyright
  * information.
  *
 ***************************************************************/

/* This HAL exists to make driver and network stack implementation
 * as seperate as possible. This will keep the network stack card-
 * agnostic, hopefully.
 *
 * Network stack code should NOT include this file. They should
 * instead include hal_stack.h and use the functions provided.
 * All the functions listed in this file must be implemented by 
 * the drivers themselves. This is mereley a list of prototypes
 * that the drivers need to implement.
 */

#ifndef AWOS_HAL_DRIVER_H
#define AWOS_HAL_DRIVER_H

#include <stdint.h>
#include <sys/types.h>

/* WARNING WARNING WARNING
 * NEVER EVER EVER directly use "delete" on a Net_IfaceDriver object.
 * this->DisposeDriver(); will delete the object if it is successful.
 *
 * Also, realise you must call this->InitializeDriver(); after creating a new
 * Net_IfaceDriver object.
 */
class Net_IfaceDriver
{
public:
	const char *name;

	/** Initializes the driver.  Returns FALSE if Iface doesn't exist
	  * and/or if Iface doesn't init properly **/
	virtual BOOL InitializeDriver() = 0;
	/** Specifies whether the driver is in use at the moment (basicaly,
	  * if it is loaded, inited, and active **/
	virtual BOOL IsDriverInUse() = 0;
	/** Attempts to dispose of the driver.  Returns FALSE if a data xfer
	  * is in progress or there is another reason the driver can't be
	  * disposed **/
	virtual BOOL DisposeDriver() = 0;

	/** Data rx/tx functions **/
	virtual int PushData(int size, void *data) = 0;
	virtual int RetrData(int size, void *data) = 0;

	/** Status functions **/
	virtual uint32_t GetDeviceFlags(void) = 0;
	virtual BOOL IsFlagSet(uint32_t) = 0;

	virtual BOOL SetDeviceFlag(uint32_t) = 0;

	Net_IfaceDriver() { };
	virtual ~Net_IfaceDriver() { };
//private: // No. This fails to work correctly when we actually want to touch stuff. -GM
	uint8_t irq;
	uint16_t basePort;
	
	unsigned char mac_addr[6];
};

/***** device flags *****/
#define NET_IFFLAG_LINKUP	0x01
#define NET_IFFLAG_SPEED_100M	0x02
#define NET_IFFLAG_DUPLEX	0x04

#endif
