/***************************************************************
 *
 * net/ifaces/e1000.h
 * Intel E1000-based Network Interface Controller Driver for ADM
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include "../hal/hal_driver.h"	// Net_IfaceDriver proto
extern "C"
{
	#include <pci.h>		// struct pcidev_ll
}

class Net_E1000IfDrv : public Net_IfaceDriver
{
public:
	Net_E1000IfDrv();
	~Net_E1000IfDrv();
	
	BOOL InitializeDriver();
	BOOL IsDriverInUse();
	BOOL DisposeDriver();

	int PushData(int, void *);
	int RetrData(int, void *);

	uint32_t GetDeviceFlags();
	BOOL IsFlagSet(uint32_t);
	BOOL SetDeviceFlag(uint32_t);
private:
	uint32_t GetDeviceStatus();	// this is the _internal_ status reg
	
	struct pcidev_ll *pciDevice;	// The PCI device object for our
					// e1000 instance.
	
	static const int DEVS_SUPP = 24;// This is the number of devices
					// currently supported.  Inc when
					// adding new ones to the list.
	
	uint16_t pciDevicesSupported[DEVS_SUPP];

	uint32_t *mem_start;
};

#define NET_E1000_REG_CTRL	0x00	// Device Control
#define NET_E1000_REG_STATUS	0x08	// Device Status
#define NET_E1000_REG_EECD	0x10	// EEPROM/Flash Control/Data
#define NET_E1000_REG_EERD	0x14	// EEPROM Read
#define NET_E1000_REG_CTRL_EXT	0x18	// Extended Device Control

