/****************************************************************

kernel/shell.c
Kludgey shell for AWOS Kernel

See LICENSE file included with this source for copyright
information.

****************************************************************/

extern "C"
{
#include <kernel.h>
#include <cons.h>
#include <input.h>
#include <string.h>
#include <sched.h>
#include "drivers/ata/ata.h"
#include "net/hal/hal_driver.h"
#include "net/hal/iface-mgr.h"
#include <stdlib.h>

#include <v86.h>

#include <fs.h>
#include <pci.h>
#include <usb.h>

extern void yield(void);
extern void exit(int);
extern AtaController *pController;

void ifconfig(void);
void hover(void);
void hdd(void);
extern void usermode(void);

void KiRestart();

extern uint32_t v86t_start, v86t_end;
};

// Basic V86 test program.
void v86(void)
{
#ifndef __PPC
	printf("Copying program\n");
	memcpy((void *)0x10000, (void *)&v86t_start, (uint32_t)&v86t_end-(uint32_t)&v86t_start);
	csleep(50);
	printf("V86 started\n");
	KiV86Enter(0x1000,0x0000,0xFFFC);
	printf("And here we are again.\n");
	//exit(0);
#endif /* !__PPC */
};

// Hex reader/writer.
// TODO: move this somewhere better
uint32_t atoih32(char *p)
{
	uint32_t ret = 0;
	
	while((*p >= '0' && *p <= '9') || ((*p | 0x20) >= 'a' && (*p | 0x20) <= 'f'))
	{
		ret <<= 4;
		if(*p >= '0' && *p <= '9')
			ret |= (*p - '0');
		else
			ret |= ((*p | 0x20) - 'a' + 0xA);
		
		p++;
	};
	return ret;
};

void hread(char *cmd)
{
	uint32_t regi, regv;
	
	regi = atoih32(cmd);
	regv = *(uint32_t *)regi;
	printf("value at %08X = %08X\n", regi, regv);
};

void hwrite8(char *cmd)
{
	uint32_t regi, regv;
	
	regi = atoih32(cmd);
	cmd = (char *)skiptosp(cmd);
	regv = atoih32(cmd);
	*(uint8_t *)regi = (uint8_t)regv;
	printf("value at %08X is now %02X\n", regi, regv);
};

void hwrite16(char *cmd)
{
	uint32_t regi, regv;
	
	regi = atoih32(cmd);
	cmd = (char *)skiptosp(cmd);
	regv = atoih32(cmd);
	*(uint16_t *)regi = (uint16_t)regv;
	printf("value at %08X is now %04X\n", regi, regv);
};

void hwrite32(char *cmd)
{
	uint32_t regi, regv;
	
	regi = atoih32(cmd);
	cmd = (char *)skiptosp(cmd);
	regv = atoih32(cmd);
	*(uint32_t *)regi = regv;
	printf("value at %08X is now %08X\n", regi, regv);
};

void ifconfig(char *cmd)
{
	Net_IfaceDriver *eth1;

	eth1 = NetHalGetInterface(1);

	if(eth1 == NULL)
	{
		printf("Error getting interface: No network controllers.\n");
		// TODO: get tasks set up properly - otherwise we end up killing the shell
		//exit(1);
		return;
	}

	if(!strcmp(cmd, "up"))
	{
		if(eth1->SetDeviceFlag(NET_IFFLAG_LINKUP))
			printf("eth1: Should be up.\n");
		else
			printf("eth1: Failed to bring up.\n");
	};

	printf("eth1\t(%s)\n", eth1->name);
	printf("\tinet addr: not supported\n");
	printf("\tinet6 addr: *really* not supported\n");
	printf("\tmac addr: %02X:%02X:%02X:%02X:%02X:%02X\n"
		,eth1->mac_addr[0]
		,eth1->mac_addr[1]
		,eth1->mac_addr[2]
		,eth1->mac_addr[3]
		,eth1->mac_addr[4]
		,eth1->mac_addr[5]
			);
	printf("\t%s %s\n", (eth1->IsFlagSet(NET_IFFLAG_LINKUP) ? "UP " : "DOWN "), (eth1->IsFlagSet(NET_IFFLAG_DUPLEX) ? "FD " : "HD "));
	//printf("\tBase address:%d Memory base:%d", eth1->io_start, eth1->mem_start);

	// TODO: get tasks set up properly - otherwise we end up killing the shell
	//exit(0);
	return;
};

// Test for various graphics primitives; you can mess with this if you want.
void gfxtest(void)
{
	int c1,c2,c3,c4;
	int x,y;
	
	x = currtty->tx;
	y = currtty->ty;
	
	switch(currtty->bpp)
	{
		case 4:
			c1 = 0xC;
			c2 = 0xA;
			c3 = 0x9;
			c4 = 0xF;
			break;
		case 8:
			c1 = 0xFC;
			c2 = 0xFA;
			c3 = 0xF9;
			c4 = 0xFF;
			break;
		case 15:
			c1 = 0x7C00;
			c2 = 0x03E0;
			c3 = 0x001F;
			c4 = 0x7FFF;
			break;
		case 16:
			c1 = 0xF800;
			c2 = 0x07E0;
			c3 = 0x001F;
			c4 = 0xFFFF;
			break;
		case 24:
		case 32:
			c1 = 0xFF0000;
			c2 = 0x00FF00;
			c3 = 0x0000FF;
			c4 = 0xFFFFFF;
			break;
		default:
			c1 = c2 = c3 = c4 = 0xFFFFFF;
			break;
	};
	
	GfxTriSolid(currtty, 50,50, 150,180, 150,100 ,c1);
	GfxTriSolid(currtty, 150,180, 150,100, 250,50 ,c2);
	GfxTriSolid(currtty, 50,50, 150,100, 250,50 ,c3);
	
	GfxLine(currtty, 50,50,250,50, c4);
	GfxLine(currtty, 250,50,150,180, c4);
	GfxLine(currtty, 150,180,50,50, c4);
	
	GfxLine(currtty, 50,50,150,100, c4);
	GfxLine(currtty, 250,50,150,100, c4);
	GfxLine(currtty, 150,180,150,100, c4);
	
	//exit(0);
}

// USB frame dumping program.
void usbdump(void)
{
	int i;
	struct usbhost_ll *h;
	
	// Check all hosts.
	for(h = usbroot; h != NULL; h = h->n)
	{
		printf("Checking host at %04X for active requests\n", h->port);
		printf("Frame pointer %08X(%08X), index %04X, status %04X, command %04X\n"
			,UsbUhciPortIn32(h, USB_UHCI_FLBASEADD)
			,(uint32_t)(h->framelist)
			,UsbUhciPortIn16(h, USB_UHCI_FRNUM) & 0x7FF
			,UsbUhciStatusRead(h)
			,UsbUhciPortIn16(h, USB_UHCI_USBCMD)
				);
		printf("Port status: Port1 %04X, Port2 %04X\n"
			,UsbUhciPortIn16(h, USB_UHCI_PORTSC1)
			,UsbUhciPortIn16(h, USB_UHCI_PORTSC2)
				);
		
		// Check for all non-terminated USB requests.
		for(i=0 ; i < 1024 ; i++)
		{
			// Is it terminated?
			if(!(h->framelist[i] & USB_UHCI_FLP_TERM))
			{
				printf("  Entry %03i is a ", i);
				if(h->framelist[i] & USB_UHCI_FLP_QH)
				{
					printf("queue head\n");
				} else {
					printf("transfer descriptor\n");
				};
			};
		};
	};
};

// Filesystem test
void fstest_iter(int tabs, struct fs_def *def, struct fs_dir *d)
{
	struct fs_dir *dc;
	int i;
	
	while(d != NULL)
	{
		for(i=0;i<tabs;i++)
			printf(" ");
		
		printf("%s", d);
		dc = def->fs_getdir(d);
		fstest_iter(tabs + 1, def, dc);
		
		dc = def->fs_getnextent(d);
		free(d);
		d = dc;
	};
};

void fstest(void)
{
	struct fs_dir *d;
	struct fs_def *def;
	
	printf("Root directory info:\n");
	def = fs_root_def;
	d = def->fs_getdir(NULL);
	fstest_iter(1, def, d);
};

void shell(void)
{
	int i,j;
	char *v;
	char cmstr[1024];
	
	printf("\nAWO$ ");
	for (i = 0;;)
	{
		j = KbdReadKey();
		if(j != 0 && j < 0x100)
		{
			cmstr[i] = j;
			if(cmstr[i] == '\x08')
			{
				if(i != 0)
				{
					i--;
					printf("\x08");
				};
			}
			else if(cmstr[i] == '\n')
			{
				printf("\n");
				cmstr[i] = '\0';
				v = &cmstr[0];
				if(!strcmpsp(v,"help")) {
					v = (char *)skiptosp(v);
					if(*v == '\0') {
						printf(
							"Commands:\n"
							"echo "
							"cls/clear "
							"ps "
							"exit\n"
								);
					}
					else if(!strcmpsp(v, "echo"))
					{
						printf("echo <text>\n\tPrint pretty text to the screen.");
					}
					else if(!strcmpsp(v, "cls") || !strcmpsp(v, "clear"))
					{
						printf("cls/clear\n\tClear the screen.");
					}
					else if(!strcmpsp(v, "ps"))
					{
						printf("ps\n\tShow a list of running processes.");
					}
					else if(!strcmpsp(v, "hover"))
					{
						printf("hover\n\tTEST: Multithreading/sleeping test.");
					}
					else if(!strcmpsp(v, "v86"))
					{
						printf("v86\n\tTEST: V86 mode.");
					}
					else if(!strcmpsp(v, "gfxtest"))
					{
						printf("gfxtest\n\tTEST: Various graphics functions.");
					}
					else if(!strcmpsp(v, "exit"))
					{
						printf("exit\n\tExits the shell.");
					}
				}
				else if(!strcmpsp(v,"echo")) {
					v = (char *)skiptosp(v);
					printf("%s\n",v);
				}
				else if(!strcmpsp(v,"cls") || !strcmpsp(v,"clear")) {
					VgaClearScreen();
				}
				else if(!strcmpsp(v,"hover")) {
					__KiCreateTask((uint32_t)&hover, (char *)"hover");
				}
				else if(!strcmpsp(v,"hdd")) {
					__KiCreateTask((uint32_t)&hdd, (char *)"hdd");
				}
				else if(!strcmpsp(v,"ifconfig")) {
					ifconfig((char *)skiptosp(v));
				}
				else if(!strcmpsp(v,"v86")) {
					v86();
				}
				else if(!strcmpsp(v,"gfxtest")) {
					gfxtest();
				}
				else if(!strcmpsp(v,"usbdump")) {
					usbdump();
				}
				else if(!strcmpsp(v,"fstest")) {
					fstest();
				}
				else if(!strcmpsp(v, "usermode")) {
					__KiCreateTaskEx((uint32_t)&usermode, (char *)"User mode test", 0x09 /* TSK_RUNNABLE, TSK_USER */, (char *)"", 0);
				}
				else if(!strcmpsp(v,"hread")) {
					hread((char *)skiptosp(v));
				}
				else if(!strcmpsp(v,"hwrite8")) {
					hwrite8((char *)skiptosp(v));
				}
				else if(!strcmpsp(v,"hwrite16")) {
					hwrite16((char *)skiptosp(v));
				}
				else if(!strcmpsp(v,"hwrite32")) {
					hwrite32((char *)skiptosp(v));
				}
				else if(!strcmpsp(v,"exit")) {
					printf("This is the last shell.  Exiting will reset the system.\nAre you sure you want to exit?[Y,N]:");
					for(;;){
						do {
							j = KbdReadKey();
						} while(j == 0);
						if ((j|0x20) == 'y')
						{
							printf("Y\n");
							yield();
							KiRestart();
							exit(0);
							break;
						}
						else if ((j|0x20) == 'n')
						{
							printf("N\n");
							break;
						};
					}
				}
				else if(!strcmpsp(v,"ps")) {
#ifndef __PPC
					task_t task;
					task = idletask;
					printf("Process list:\n");
					do {
						printf("- %05i - %s\n", task->pid, task->name);
						task = task->next;
					} while(task != idletask);
#endif /* !__PPC */
				} else {
					if(!strcmp(v, "awos sucks")) {
						printf("Quit whining, HellDragon.\n");
					} else if(i != 0) {
						printf("That doesn't look much like a command.\n");
					}
				}
				printf("\nAWO$ ");
				i = 0;
			} else {
				printf("%c",(char)j);
				i++;
			};
		};
		if(j == 0)
			yield();
	};
	
	while(1) asm("hlt");
	exit(0);
};

// Scheduler spawning test.
void hover(void)
{
	printf("hover: Hovering for 500 clock cycles.\n");
	csleep(500);
	printf("Hovering done.\n");
	exit(0);
};

// Hard disk driver test.
void hdd(void)
{
	if(pController == NULL)
	{
		printf("Error: You did not start AWOS with the 'hdd' parameter, or there was an error in driver registration.\n");
		exit(1);
	}
	
	printf("Obtaining disk...");
	AtaDisk hDisk = AtaObtainDisk(pController, 0);
	printf("%08X\n", hDisk);
	unsigned char *mbr;
	mbr = (unsigned char *)AtaRead(hDisk, TRUE, 1, 0);
	printf("pointer at %08X last two bytes of MBR: %02X %02X\n", mbr, mbr[510], mbr[511]);
	printf("Releasing disk...");
	AtaReleaseDisk(hDisk);
	printf("done\n");
	exit(0);
};
